@@ -76,7 +76,7 @@ def _grid_in_property(
     if read_only:
         docstring += "\n\nThis attribute is read-only."
     elif closed_only:
-        docstring = "%s\n\n%s" % (
+        docstring = "{}\n\n{}".format(
             docstring,
             "This attribute is read-only and "
             "can only be read after :meth:`close` "
@@ -114,7 +114,7 @@ def _disallow_transactions(session: Optional[ClientSession]) -> None:
         raise InvalidOperation("GridFS does not support multi-document transactions")
 
 
-class GridIn(object):
+class GridIn:
     """Class to write data to GridFS."""
 
     def __init__(
@@ -292,7 +292,7 @@ class GridIn(object):
             self.__flush_buffer()
             # The GridFS spec says length SHOULD be an Int64.
             self._file["length"] = Int64(self._position)
-            self._file["uploadDate"] = datetime.datetime.utcnow()
+            self._file["uploadDate"] = datetime.datetime.now(tz=datetime.timezone.utc)
 
             return self._coll.files.insert_one(self._file, session=self._session)
         except DuplicateKeyError:
@@ -381,7 +381,7 @@ class GridIn(object):
     def writelines(self, sequence: Iterable[Any]) -> None:
         """Write a sequence of strings to the file.
 
-        Does not add seperators.
+        Does not add separators.
         """
         for line in sequence:
             self.write(line)
@@ -463,7 +463,10 @@ class GridOut(io.IOBase):
         self.__files = root_collection.files
         self.__file_id = file_id
         self.__buffer = EMPTY
+        # Start position within the current buffered chunk.
+        self.__buffer_pos = 0
         self.__chunk_iter = None
+        # Position within the total file.
         self.__position = 0
         self._file = file_document
         self._session = session
@@ -494,7 +497,7 @@ class GridOut(io.IOBase):
             self._file = self.__files.find_one({"_id": self.__file_id}, session=self._session)
             if not self._file:
                 raise NoFile(
-                    "no file in gridfs collection %r with _id %r" % (self.__files, self.__file_id)
+                    f"no file in gridfs collection {self.__files!r} with _id {self.__file_id!r}"
                 )
 
     def __getattr__(self, name: str) -> Any:
@@ -510,12 +513,12 @@ class GridOut(io.IOBase):
         """Reads a chunk at a time. If the current position is within a
         chunk the remainder of the chunk is returned.
         """
-        received = len(self.__buffer)
+        received = len(self.__buffer) - self.__buffer_pos
         chunk_data = EMPTY
         chunk_size = int(self.chunk_size)
 
         if received > 0:
-            chunk_data = self.__buffer
+            chunk_data = self.__buffer[self.__buffer_pos :]
         elif self.__position < int(self.length):
             chunk_number = int((received + self.__position) / chunk_size)
             if self.__chunk_iter is None:
@@ -531,8 +534,60 @@ class GridOut(io.IOBase):
 
         self.__position += len(chunk_data)
         self.__buffer = EMPTY
+        self.__buffer_pos = 0
         return chunk_data
 
+    def _read_size_or_line(self, size: int = -1, line: bool = False) -> bytes:
+        """Internal read() and readline() helper."""
+        self._ensure_file()
+        remainder = int(self.length) - self.__position
+        if size < 0 or size > remainder:
+            size = remainder
+
+        if size == 0:
+            return EMPTY
+
+        received = 0
+        data = []
+        while received < size:
+            needed = size - received
+            if self.__buffer:
+                # Optimization: Read the buffer with zero byte copies.
+                buf = self.__buffer
+                chunk_start = self.__buffer_pos
+                chunk_data = memoryview(buf)[self.__buffer_pos :]
+                self.__buffer = EMPTY
+                self.__buffer_pos = 0
+                self.__position += len(chunk_data)
+            else:
+                buf = self.readchunk()
+                chunk_start = 0
+                chunk_data = memoryview(buf)
+            if line:
+                pos = buf.find(NEWLN, chunk_start, chunk_start + needed) - chunk_start
+                if pos >= 0:
+                    # Decrease size to exit the loop.
+                    size = received + pos + 1
+                    needed = pos + 1
+            if len(chunk_data) > needed:
+                data.append(chunk_data[:needed])
+                # Optimization: Save the buffer with zero byte copies.
+                self.__buffer = buf
+                self.__buffer_pos = chunk_start + needed
+                self.__position -= len(self.__buffer) - self.__buffer_pos
+            else:
+                data.append(chunk_data)
+            received += len(chunk_data)
+
+        # Detect extra chunks after reading the entire file.
+        if size == remainder and self.__chunk_iter:
+            try:
+                self.__chunk_iter.next()
+            except StopIteration:
+                pass
+
+        return b"".join(data)
+
     def read(self, size: int = -1) -> bytes:
         """Read at most `size` bytes from the file (less if there
         isn't enough data).
@@ -548,36 +603,7 @@ class GridOut(io.IOBase):
            entire file. Previously, this method would check for extra chunks
            on every call.
         """
-        self._ensure_file()
-
-        remainder = int(self.length) - self.__position
-        if size < 0 or size > remainder:
-            size = remainder
-
-        if size == 0:
-            return EMPTY
-
-        received = 0
-        data = io.BytesIO()
-        while received < size:
-            chunk_data = self.readchunk()
-            received += len(chunk_data)
-            data.write(chunk_data)
-
-        # Detect extra chunks after reading the entire file.
-        if size == remainder and self.__chunk_iter:
-            try:
-                self.__chunk_iter.next()
-            except StopIteration:
-                pass
-
-        self.__position -= received - size
-
-        # Return 'size' bytes and store the rest.
-        data.seek(size)
-        self.__buffer = data.read()
-        data.seek(0)
-        return data.read(size)
+        return self._read_size_or_line(size=size)
 
     def readline(self, size: int = -1) -> bytes:  # type: ignore[override]
         """Read one line or up to `size` bytes from the file.
@@ -585,33 +611,7 @@ class GridOut(io.IOBase):
         :Parameters:
          - `size` (optional): the maximum number of bytes to read
         """
-        remainder = int(self.length) - self.__position
-        if size < 0 or size > remainder:
-            size = remainder
-
-        if size == 0:
-            return EMPTY
-
-        received = 0
-        data = io.BytesIO()
-        while received < size:
-            chunk_data = self.readchunk()
-            pos = chunk_data.find(NEWLN, 0, size)
-            if pos != -1:
-                size = received + pos + 1
-
-            received += len(chunk_data)
-            data.write(chunk_data)
-            if pos != -1:
-                break
-
-        self.__position -= received - size
-
-        # Return 'size' bytes and store the rest.
-        data.seek(size)
-        self.__buffer = data.read()
-        data.seek(0)
-        return data.read(size)
+        return self._read_size_or_line(size=size, line=True)
 
     def tell(self) -> int:
         """Return the current position of this file."""
@@ -640,10 +640,10 @@ class GridOut(io.IOBase):
         elif whence == _SEEK_END:
             new_pos = int(self.length) + pos
         else:
-            raise IOError(22, "Invalid value for `whence`")
+            raise OSError(22, "Invalid value for `whence`")
 
         if new_pos < 0:
-            raise IOError(22, "Invalid value for `pos` - must be positive")
+            raise OSError(22, "Invalid value for `pos` - must be positive")
 
         # Optimization, continue using the same buffer and chunk iterator.
         if new_pos == self.__position:
@@ -651,6 +651,7 @@ class GridOut(io.IOBase):
 
         self.__position = new_pos
         self.__buffer = EMPTY
+        self.__buffer_pos = 0
         if self.__chunk_iter:
             self.__chunk_iter.close()
             self.__chunk_iter = None
@@ -731,7 +732,7 @@ class GridOut(io.IOBase):
         pass
 
 
-class _GridOutChunkIterator(object):
+class _GridOutChunkIterator:
     """Iterates over a file's chunks using a single cursor.
 
     Raises CorruptGridFile when encountering any truncated, missing, or extra
@@ -831,7 +832,7 @@ class _GridOutChunkIterator(object):
             self._cursor = None
 
 
-class GridOutIterator(object):
+class GridOutIterator:
     def __init__(self, grid_out: GridOut, chunks: Collection, session: ClientSession):
         self.__chunk_iter = _GridOutChunkIterator(grid_out, chunks, session, 0)
 
@@ -877,7 +878,7 @@ class GridOutCursor(Cursor):
         # Hold on to the base "fs" collection to create GridOut objects later.
         self.__root_collection = collection
 
-        super(GridOutCursor, self).__init__(
+        super().__init__(
             collection.files,
             filter,
             skip=skip,
@@ -891,7 +892,7 @@ class GridOutCursor(Cursor):
     def next(self) -> GridOut:
         """Get next GridOut object from cursor."""
         _disallow_transactions(self.session)
-        next_file = super(GridOutCursor, self).next()
+        next_file = super().next()
         return GridOut(self.__root_collection, file_document=next_file, session=self.session)
 
     __next__ = next
