@@ -27,9 +27,10 @@ access:
   >>> c = MongoClient()
   >>> c.test_database
   Database(MongoClient(host=['localhost:27017'], document_class=dict, tz_aware=False, connect=True), 'test_database')
-  >>> c['test-database']
+  >>> c["test-database"]
   Database(MongoClient(host=['localhost:27017'], document_class=dict, tz_aware=False, connect=True), 'test-database')
 """
+from __future__ import annotations
 
 import contextlib
 import os
@@ -53,7 +54,8 @@ from typing import (
     cast,
 )
 
-from bson.codec_options import DEFAULT_CODEC_OPTIONS, CodecOptions, TypeRegistry
+import bson
+from bson.codec_options import DEFAULT_CODEC_OPTIONS, TypeRegistry
 from bson.son import SON
 from bson.timestamp import Timestamp
 from pymongo import (
@@ -90,7 +92,13 @@ from pymongo.server_type import SERVER_TYPE
 from pymongo.settings import TopologySettings
 from pymongo.topology import Topology, _ErrorContext
 from pymongo.topology_description import TOPOLOGY_TYPE, TopologyDescription
-from pymongo.typings import _Address, _CollationIn, _DocumentType, _Pipeline
+from pymongo.typings import (
+    _Address,
+    _CollationIn,
+    _DocumentType,
+    _DocumentTypeArg,
+    _Pipeline,
+)
 from pymongo.uri_parser import (
     _check_options,
     _handle_option_deprecations,
@@ -394,6 +402,10 @@ class MongoClient(common.BaseObject, Generic[_DocumentType]):
             "mongodb+srv://" URIs. Defaults to "mongodb". Use it like so::
 
                 MongoClient("mongodb+srv://example.com/?srvServiceName=customname")
+          - `srvMaxHosts`: (int) limits the number of mongos-like hosts a client will
+            connect to. More specifically, when a "mongodb+srv://" connection string
+            resolves to more than srvMaxHosts number of hosts, the client will randomly
+            choose an srvMaxHosts sized subset of hosts.
 
 
           | **Write Concern options:**
@@ -568,8 +580,8 @@ class MongoClient(common.BaseObject, Generic[_DocumentType]):
                keyword arguments.
              - The default for `uuidRepresentation` was changed from
                ``pythonLegacy`` to ``unspecified``.
-             - Added the ``srvServiceName`` and ``maxConnecting`` URI and
-               keyword argument.
+             - Added the ``srvServiceName``, ``maxConnecting``, and ``srvMaxHosts`` URI and
+               keyword arguments.
 
         .. versionchanged:: 3.12
            Added the ``server_api`` keyword argument.
@@ -794,7 +806,7 @@ class MongoClient(common.BaseObject, Generic[_DocumentType]):
         self.__kill_cursors_queue: List = []
 
         self._event_listeners = options.pool_options._event_listeners
-        super(MongoClient, self).__init__(
+        super().__init__(
             options.codec_options,
             options.read_preference,
             options.write_concern,
@@ -924,14 +936,13 @@ class MongoClient(common.BaseObject, Generic[_DocumentType]):
         .. code-block:: python
 
             try:
-                with client.watch(
-                        [{'$match': {'operationType': 'insert'}}]) as stream:
+                with client.watch([{"$match": {"operationType": "insert"}}]) as stream:
                     for insert_change in stream:
                         print(insert_change)
             except pymongo.errors.PyMongoError:
                 # The ChangeStream encountered an unrecoverable error or the
                 # resume attempt failed to recreate the cursor.
-                logging.error('...')
+                logging.error("...")
 
         For a precise description of the resume process see the
         `change streams specification`_.
@@ -1408,12 +1419,18 @@ class MongoClient(common.BaseObject, Generic[_DocumentType]):
                 if retryable_error:
                     session._unpin()
                 if not retryable_error or (is_retrying() and not multiple_retries):
-                    raise
+                    if exc.has_error_label("NoWritesPerformed") and last_error:
+                        raise last_error from exc
+                    else:
+                        raise
                 if bulk:
                     bulk.retrying = True
                 else:
                     retrying = True
-                last_error = exc
+                if not exc.has_error_label("NoWritesPerformed"):
+                    last_error = exc
+                if last_error is None:
+                    last_error = exc
 
     @_csot.apply
     def _retryable_read(self, func, read_pref, session, address=None, retryable=True):
@@ -1493,11 +1510,11 @@ class MongoClient(common.BaseObject, Generic[_DocumentType]):
                 if value is dict:
                     return "document_class=dict"
                 else:
-                    return "document_class=%s.%s" % (value.__module__, value.__name__)
+                    return f"document_class={value.__module__}.{value.__name__}"
             if option in common.TIMEOUT_OPTIONS and value is not None:
-                return "%s=%s" % (option, int(value * 1000))
+                return f"{option}={int(value * 1000)}"
 
-            return "%s=%r" % (option, value)
+            return f"{option}={value!r}"
 
         # Host first...
         options = [
@@ -1520,7 +1537,7 @@ class MongoClient(common.BaseObject, Generic[_DocumentType]):
         return ", ".join(options)
 
     def __repr__(self):
-        return "MongoClient(%s)" % (self._repr_helper(),)
+        return f"MongoClient({self._repr_helper()})"
 
     def __getattr__(self, name: str) -> database.Database[_DocumentType]:
         """Get a database by name.
@@ -1533,8 +1550,8 @@ class MongoClient(common.BaseObject, Generic[_DocumentType]):
         """
         if name.startswith("_"):
             raise AttributeError(
-                "MongoClient has no attribute %r. To access the %s"
-                " database, use client[%r]." % (name, name, name)
+                "MongoClient has no attribute {!r}. To access the {}"
+                " database, use client[{!r}].".format(name, name, name)
             )
         return self.__getitem__(name)
 
@@ -1669,7 +1686,8 @@ class MongoClient(common.BaseObject, Generic[_DocumentType]):
     # This method is run periodically by a background thread.
     def _process_periodic_tasks(self):
         """Process any pending kill cursors requests and
-        maintain connection pool parameters."""
+        maintain connection pool parameters.
+        """
         try:
             self._process_kill_cursors()
             self._topology.update_pool()
@@ -1726,7 +1744,7 @@ class MongoClient(common.BaseObject, Generic[_DocumentType]):
     def _return_server_session(self, server_session, lock):
         """Internal: return a _ServerSession to the pool."""
         if isinstance(server_session, _EmptyServerSession):
-            return
+            return None
         return self._topology.return_server_session(server_session, lock)
 
     def _ensure_session(self, session=None):
@@ -1745,7 +1763,7 @@ class MongoClient(common.BaseObject, Generic[_DocumentType]):
     @contextlib.contextmanager
     def _tmp_session(
         self, session: Optional[client_session.ClientSession], close: bool = True
-    ) -> "Generator[Optional[client_session.ClientSession], None, None]":
+    ) -> Generator[Optional[client_session.ClientSession], None, None]:
         """If provided session is None, lend a temporary session."""
         if session is not None:
             if not isinstance(session, client_session.ClientSession):
@@ -1869,7 +1887,7 @@ class MongoClient(common.BaseObject, Generic[_DocumentType]):
     @_csot.apply
     def drop_database(
         self,
-        name_or_database: Union[str, database.Database],
+        name_or_database: Union[str, database.Database[_DocumentTypeArg]],
         session: Optional[client_session.ClientSession] = None,
         comment: Optional[Any] = None,
     ) -> None:
@@ -1922,10 +1940,10 @@ class MongoClient(common.BaseObject, Generic[_DocumentType]):
     def get_default_database(
         self,
         default: Optional[str] = None,
-        codec_options: Optional[CodecOptions] = None,
+        codec_options: Optional[bson.CodecOptions[_DocumentTypeArg]] = None,
         read_preference: Optional[_ServerMode] = None,
         write_concern: Optional[WriteConcern] = None,
-        read_concern: Optional["ReadConcern"] = None,
+        read_concern: Optional[ReadConcern] = None,
     ) -> database.Database[_DocumentType]:
         """Get the database named in the MongoDB connection URI.
 
@@ -1983,10 +2001,10 @@ class MongoClient(common.BaseObject, Generic[_DocumentType]):
     def get_database(
         self,
         name: Optional[str] = None,
-        codec_options: Optional[CodecOptions] = None,
+        codec_options: Optional[bson.CodecOptions[_DocumentTypeArg]] = None,
         read_preference: Optional[_ServerMode] = None,
         write_concern: Optional[WriteConcern] = None,
-        read_concern: Optional["ReadConcern"] = None,
+        read_concern: Optional[ReadConcern] = None,
     ) -> database.Database[_DocumentType]:
         """Get a :class:`~pymongo.database.Database` with the given name and
         options.
@@ -2105,7 +2123,7 @@ def _add_retryable_write_error(exc, max_wire_version):
         exc._add_error_label("RetryableWriteError")
 
 
-class _MongoClientErrorHandler(object):
+class _MongoClientErrorHandler:
     """Handle errors raised when executing an operation."""
 
     __slots__ = (
