@@ -17,7 +17,6 @@
 import os
 import queue
 import random
-import threading
 import time
 import warnings
 import weakref
@@ -37,6 +36,7 @@ from pymongo.errors import (
     WriteError,
 )
 from pymongo.hello import Hello
+from pymongo.lock import _create_lock
 from pymongo.monitor import SrvMonitor
 from pymongo.pool import PoolOptions
 from pymongo.server import Server
@@ -75,7 +75,7 @@ def process_events_queue(queue_ref):
     return True  # Continue PeriodicExecutor.
 
 
-class Topology(object):
+class Topology:
     """Monitor a topology of one or more servers."""
 
     def __init__(self, topology_settings):
@@ -127,7 +127,7 @@ class Topology(object):
         self._seed_addresses = list(topology_description.server_descriptions())
         self._opened = False
         self._closed = False
-        self._lock = threading.Lock()
+        self._lock = _create_lock()
         self._condition = self._settings.condition_class(self._lock)
         self._servers = {}
         self._pid = None
@@ -175,8 +175,8 @@ class Topology(object):
         elif pid != self._pid:
             self._pid = pid
             warnings.warn(
-                "MongoClient opened before fork. Create MongoClient only "
-                "after forking. See PyMongo's documentation for details: "
+                "MongoClient opened before fork. May not be entirely fork-safe, "
+                "proceed with caution. See PyMongo's documentation for details: "
                 "https://pymongo.readthedocs.io/en/stable/faq.html#"
                 "is-pymongo-fork-safe"
             )
@@ -236,8 +236,7 @@ class Topology(object):
             # No suitable servers.
             if timeout == 0 or now > end_time:
                 raise ServerSelectionTimeoutError(
-                    "%s, Timeout: %ss, Topology Description: %r"
-                    % (self._error_message(selector), timeout, self.description)
+                    f"{self._error_message(selector)}, Timeout: {timeout}s, Topology Description: {self.description!r}"
                 )
 
             self._ensure_opened()
@@ -271,7 +270,7 @@ class Topology(object):
         """Like select_servers, but choose a random server if several match."""
         server = self._select_server(selector, server_selection_timeout, address)
         if _csot.get_timeout():
-            _csot.set_rtt(server.description.round_trip_time)
+            _csot.set_rtt(server.description.min_round_trip_time)
         return server
 
     def select_server_by_address(self, address, server_selection_timeout=None):
@@ -431,7 +430,7 @@ class Topology(object):
             ):
                 return set()
 
-            return set([sd.address for sd in selector(self._new_selection())])
+            return {sd.address for sd in selector(self._new_selection())}
 
     def get_secondaries(self):
         """Return set of secondary addresses."""
@@ -499,7 +498,8 @@ class Topology(object):
     def close(self):
         """Clear pools and terminate monitors. Topology does not reopen on
         demand. Any further operations will raise
-        :exc:`~.errors.InvalidOperation`."""
+        :exc:`~.errors.InvalidOperation`.
+        """
         with self._lock:
             for server in self._servers.values():
                 server.close()
@@ -807,14 +807,14 @@ class Topology(object):
                 else:
                     return "No %s available for writes" % server_plural
             else:
-                return 'No %s match selector "%s"' % (server_plural, selector)
+                return f'No {server_plural} match selector "{selector}"'
         else:
             addresses = list(self._description.server_descriptions())
             servers = list(self._description.server_descriptions().values())
             if not servers:
                 if is_replica_set:
                     # We removed all servers because of the wrong setName?
-                    return 'No %s available for replica set name "%s"' % (
+                    return 'No {} available for replica set name "{}"'.format(
                         server_plural,
                         self._settings.replica_set_name,
                     )
@@ -844,7 +844,7 @@ class Topology(object):
         msg = ""
         if not self._opened:
             msg = "CLOSED "
-        return "<%s %s%r>" % (self.__class__.__name__, msg, self._description)
+        return f"<{self.__class__.__name__} {msg}{self._description!r}>"
 
     def eq_props(self):
         """The properties to use for MongoClient/Topology equality checks."""
@@ -860,7 +860,7 @@ class Topology(object):
         return hash(self.eq_props())
 
 
-class _ErrorContext(object):
+class _ErrorContext:
     """An error with context for SDAM error handling."""
 
     def __init__(self, error, max_wire_version, sock_generation, completed_handshake, service_id):
